 /*
 *    PIC C Compiler (CCS)
 *
 *    @section Author
 *
 *    microchipotle.com
 *
 */
 
/*
Sparkfun's ADXL345 Library Main Source File
SparkFun_ADXL345.cpp
E.Robert @ SparkFun Electronics
Created: Jul 13, 2016
Updated: Sep 06, 2016
Modified Bildr ADXL345 Source File @ http://code.bildr.org/download/959.zip
to support both I2C 
Hardware Resources:
- Arduino Development Board
- SparkFun Triple Access Accelerometer ADXL345
Development Environment Specifics:
Arduino 1.6.8
SparkFun Triple Axis Accelerometer Breakout - ADXL345
Arduino Uno
*/

#include <math.h>
#include <stdlib.h>
#include <stdint.h>
/*************************** REGISTER MAP ***************************/
#define ADXL345_DEVID         0x00      // Device ID
#define ADXL345_RESERVED1     0x01      // Reserved. Do Not Access. 
#define ADXL345_THRESH_TAP    0x1D      // Tap Threshold. 
#define ADXL345_OFSX          0x1E      // X-Axis Offset. 
#define ADXL345_OFSY          0x1F      // Y-Axis Offset.
#define ADXL345_OFSZ          0x20      // Z- Axis Offset.
#define ADXL345_DUR           0x21      // Tap Duration.
#define ADXL345_LATENT        0x22      // Tap Latency.
#define ADXL345_WINDOW        0x23      // Tap Window.
#define ADXL345_THRESH_ACT    0x24      // Activity Threshold
#define ADXL345_THRESH_INACT  0x25      // Inactivity Threshold
#define ADXL345_TIME_INACT    0x26      // Inactivity Time
#define ADXL345_ACT_INACT_CTL 0x27      // Axis Enable Control for Activity and Inactivity Detection
#define ADXL345_THRESH_FF     0x28      // Free-Fall Threshold.
#define ADXL345_TIME_FF       0x29      // Free-Fall Time.
#define ADXL345_TAP_AXES      0x2A      // Axis Control for Tap/Double Tap.
#define ADXL345_ACT_TAP_STATUS   0x2B      // Source of Tap/Double Tap
#define ADXL345_BW_RATE          0x2C      // Data Rate and Power mode Control
#define ADXL345_POWER_CTL        0x2D      // Power-Saving Features Control
#define ADXL345_INT_ENABLE       0x2E      // Interrupt Enable Control
#define ADXL345_INT_MAP          0x2F      // Interrupt Mapping Control
#define ADXL345_INT_SOURCE       0x30      // Source of Interrupts
#define ADXL345_DATA_FORMAT      0x31      // Data Format Control
#define ADXL345_DATAX0        0x32      // X-Axis Data 0
#define ADXL345_DATAX1        0x33      // X-Axis Data 1
#define ADXL345_DATAY0        0x34      // Y-Axis Data 0
#define ADXL345_DATAY1        0x35      // Y-Axis Data 1
#define ADXL345_DATAZ0        0x36      // Z-Axis Data 0
#define ADXL345_DATAZ1        0x37      // Z-Axis Data 1
#define ADXL345_FIFO_CTL      0x38      // FIFO Control
#define ADXL345_FIFO_STATUS   0x39      // FIFO Status

#define ADXL345_BW_1600       0xF         // 1111      IDD = 40uA
#define ADXL345_BW_800        0xE         // 1110      IDD = 90uA
#define ADXL345_BW_400        0xD         // 1101      IDD = 140uA
#define ADXL345_BW_200        0xC         // 1100      IDD = 140uA
#define ADXL345_BW_100        0xB         // 1011      IDD = 140uA 
#define ADXL345_BW_50         0xA         // 1010      IDD = 140uA
#define ADXL345_BW_25         0x9         // 1001      IDD = 90uA
#define ADXL345_BW_12_5       0x8         // 1000      IDD = 60uA 
#define ADXL345_BW_6_25       0x7         // 0111      IDD = 50uA
#define ADXL345_BW_3_13       0x6         // 0110      IDD = 45uA
#define ADXL345_BW_1_56       0x5         // 0101      IDD = 40uA
#define ADXL345_BW_0_78       0x4         // 0100      IDD = 34uA
#define ADXL345_BW_0_39       0x3         // 0011      IDD = 23uA
#define ADXL345_BW_0_20       0x2         // 0010      IDD = 23uA
#define ADXL345_BW_0_10       0x1         // 0001      IDD = 23uA
#define ADXL345_BW_0_05       0x0         // 0000      IDD = 23uA


 /************************** INTERRUPT PINS **************************/
#define ADXL345_INT1_PIN      0x00      //INT1: 0
#define ADXL345_INT2_PIN      0x01      //INT2: 1


 /********************** INTERRUPT BIT POSITION **********************/
#define ADXL345_INT_DATA_READY_BIT     0x07
#define ADXL345_INT_SINGLE_TAP_BIT     0x06
#define ADXL345_INT_DOUBLE_TAP_BIT     0x05
#define ADXL345_INT_ACTIVITY_BIT       0x04
#define ADXL345_INT_INACTIVITY_BIT     0x03
#define ADXL345_INT_FREE_FALL_BIT      0x02
#define ADXL345_INT_WATERMARK_BIT      0x01
#define ADXL345_INT_OVERRUNY_BIT       0x00

#define ADXL345_DATA_READY    0x07
#define ADXL345_SINGLE_TAP    0x06
#define ADXL345_DOUBLE_TAP    0x05
#define ADXL345_ACTIVITY      0x04
#define ADXL345_INACTIVITY    0x03
#define ADXL345_FREE_FALL     0x02
#define ADXL345_WATERMARK     0x01
#define ADXL345_OVERRUNY      0x00


 /****************************** ERRORS ******************************/
#define ADXL345_OK         1      // No Error
#define ADXL345_ERROR      0      // Error Exists

#define ADXL345_NO_ERROR      0      // Initial State
#define ADXL345_READ_ERROR    1      // Accelerometer Reading Error
#define ADXL345_BAD_ARG       2      // Bad Argument

/****************************** ADDRESS ******************************/
#define ADXL345_COMMAND_BIT     (0x80) /**< Command bit **/
#define ADXL345_DEVICE           (0x53)    // Device Address for ADXL345
#define ADXL345_TO_READ          (6)      // Number of Bytes Read - Two Bytes Per Axis

   boolean status;         // Set When Error Exists 
   byte error_code;        // Initial State
   double gains[3];        // Counts to Gs
   
   void ADXL345_init();
   void powerOn(); 
   void readAccel(int16_t *xyx);
   void readG(float *x, float *y, float *z);
   void readAccel(int16_t* x, int16_t* y, int16_t* z);
   void get_Gxyz(double *xyz);
  
   void setTapThreshold(int16_t tapThreshold);
   int16_t getTapThreshold();
   void setAxisGains(double *_gains);
   void getAxisGains(double *_gains);
   void setAxisOffset(int16_t x, int16_t y, int16_t z);
   void getAxisOffset(int16_t* x, int16_t* y, int16_t*z);
   void setTapDuration(int16_t tapDuration);
   int16_t getTapDuration();
   void setDoubleTapLatency(int16_t doubleTapLatency);
   int16_t getDoubleTapLatency();
   void setDoubleTapWindow(int16_t doubleTapWindow);
   int16_t getDoubleTapWindow();
   void setActivityThreshold(int16_t activityThreshold);
   int16_t getActivityThreshold();
   void setInactivityThreshold(int16_t inactivityThreshold);
   int16_t getInactivityThreshold();
   void setTimeInactivity(int16_t timeInactivity);
   int16_t getTimeInactivity();
   void setFreeFallThreshold(int16_t freeFallthreshold);
   int16_t getFreeFallThreshold();
   void setFreeFallDuration(int16_t freeFallDuration);
   int16_t getFreeFallDuration();
   
   boolean isActivityXEnabled();
   boolean isActivityYEnabled();
   boolean isActivityZEnabled();
   boolean isInactivityXEnabled();
   boolean isInactivityYEnabled();
   boolean isInactivityZEnabled();
   boolean isActivityAc();
   boolean isInactivityAc();
   void setActivityAc(boolean state);
   void setInactivityAc(boolean state);
   
   boolean getSuppressBit();
   void setSuppressBit(boolean state);
   boolean isTapDetectionOnX();
   void setTapDetectionOnX(boolean state);
   boolean isTapDetectionOnY();
   void setTapDetectionOnY(boolean state);
   boolean isTapDetectionOnZ();
   void setTapDetectionOnZ(boolean state);
   void setTapDetectionOnXYZ(boolean stateX, boolean stateY, boolean stateZ);
   
   void setActivityX(boolean state);
   void setActivityY(boolean state);
   void setActivityZ(boolean state);
   void setActivityXYZ(boolean stateX, boolean stateY, boolean stateZ);
   void setInactivityX(boolean state);
   void setInactivityY(boolean state);
   void setInactivityZ(boolean state);
   void setInactivityXYZ(boolean stateX, boolean stateY, boolean stateZ);
   
   boolean isActivitySourceOnX();
   boolean isActivitySourceOnY();
   boolean isActivitySourceOnZ();
   boolean isTapSourceOnX();
   boolean isTapSourceOnY();
   boolean isTapSourceOnZ();
   boolean isAsleep();
   
   boolean isLowPower();
   void setLowPower(boolean state);
   double getRate();
   void setRate(double rate);
   void set_bw(int8_t bw_code);
   byte get_bw_code();  
   
   boolean triggered(byte interrupts, int16_t mask);
   
   byte getInterruptSource();
   boolean getInterruptSource(byte interruptBit);
   boolean getInterruptMapping(byte interruptBit);
   void setInterruptMapping(byte interruptBit, boolean interruptPin);
   boolean isInterruptEnabled(byte interruptBit);
   void setInterrupt(byte interruptBit, boolean state);
   void setImportantInterruptMapping(int16_t single_tap, int16_t double_tap, int16_t free_fall, int16_t activity, int16_t inactivity);
   void InactivityINT(boolean status);
   void ActivityINT(boolean status);
   void FreeFallINT(boolean status);
   void doubleTapINT(boolean status);
   void singleTapINT(boolean status);
   
   void getRangeSetting(byte* rangeSetting);
   void setRangeSetting(int16_t val);
   boolean getSelfTestBit();
   void setSelfTestBit(boolean selfTestBit);


   boolean getInterruptLevelBit();
   void setInterruptLevelBit(boolean interruptLevelBit);
   boolean getFullResBit();
   void setFullResBit(boolean fullResBit);
   boolean getJustifyBit();
   void setJustifyBit(boolean justifyBit);
   void printAllRegister();
   
   void writeTo(byte address, byte val);
   void writeToI2C(byte address, byte val);

   void readFrom(byte address, int8_t num, byte buff[]);
   void readFromI2C(byte address, int8_t num, byte buff[]);

   void setRegisterBit(byte regAdress, int16_t bitPos, boolean state);
   boolean getRegisterBit(byte regAdress, int16_t bitPos);  
   byte _buff[6] ;      //   6 Bytes Buffer
   boolean I2C = true;


int16_t constrain(int16_t value, int16_t min, int16_t max) {
    if (value < min) {
        return min;
    }
    if (value > max) {
        return max;
    }
    return value;
}

void ADXL345_init() {
   status = ADXL345_OK;
   error_code = ADXL345_NO_ERROR;

   gains[0] = 0.00376390;      // Original gain 0.00376390
   gains[1] = 0.00376009;      // Original gain 0.00376009
   gains[2] = 0.00349265;      // Original gain 0.00349265
   I2C = true;
   powerOn();
}

void powerOn() {
   //ADXL345 TURN ON
   writeTo(ADXL345_POWER_CTL, 0);   // Wakeup
   writeTo(ADXL345_POWER_CTL, 16);   // Auto_Sleep
   writeTo(ADXL345_POWER_CTL, 8);   // Measure
}

/*********************** READING G FORCE ***********************/
/*    Reads Acceleration into Three Variables:  x, y and z          */
void readG(float *x, float *y, float *z) 
{
   readFrom(ADXL345_DATAX0, ADXL345_TO_READ, _buff);   // Read Accel Data from ADXL345
   // Each Axis @ All g Ranges: 10 Bit Resolution (2 Bytes)
   *x = (int16_t)((((int16_t)_buff[1]) << 8) | ((int16_t)(_buff[0]) & 0xFF));
   *y = (int16_t)((((int16_t)_buff[3]) << 8) | ((int16_t)(_buff[2]) & 0xFF));
   *z = (int16_t)((((int16_t)_buff[5]) << 8) | ((int16_t)(_buff[4]) & 0xFF));
   
   *x *= 0.00376390;
   *y *= 0.00376009;
   *z *= 0.00349265;
}

/*********************** READING ACCELERATION ***********************/
void readAccel(int16_t *x, int16_t *y, int16_t *z) {
   readFrom(ADXL345_DATAX0, ADXL345_TO_READ, _buff);   // Read Accel Data from ADXL345
   // Each Axis @ All g Ranges: 10 Bit Resolution (2 Bytes)
   *x = (int16_t)((((int16_t)_buff[1]) << 8) | ((int16_t)(_buff[0]) & 0xFF));
   *y = (int16_t)((((int16_t)_buff[3]) << 8) | ((int16_t)(_buff[2]) & 0xFF));
   *z = (int16_t)((((int16_t)_buff[5]) << 8) | ((int16_t)(_buff[4]) & 0xFF));
}


void readAccel(int16_t *xyz){
   readAccel(xyz, xyz + 1, xyz + 2);
}

void get_Gxyz(double *xyz){
   int16_t i;
   int16 xyz_int[3];
   readAccel(xyz_int);
   for(i=0; i<3; i++){
      xyz[i] = xyz_int[i] * gains[i];
   }
}


/***************** WRITES VALUE TO ADDRESS REGISTER *****************/
void writeTo(byte address, byte val) {
      writeToI2C(address, val);
}

/************************ READING NUM BYTES *************************/
/*    Reads Num Bytes. Starts from Address Reg to _buff Array        */

void readFrom(byte address, int8_t num, byte _buff[]) {
      readFromI2C(address, num, _buff);   // If I2C Communication
}



/*************************** WRITE TO I2C ***************************/
/*      Start; Send Register Address; Send Value To Write; End      */
void writeToI2C(byte _address, byte _val) {      
  int8 buffer[2];
  buffer[0]=(int8)(ADXL345_COMMAND_BIT | _address);
  buffer[1]=_val; 
  i2c_transfer(ADXL345_stream, ADXL345_DEVICE<<1, buffer, 2); 
}



void readFromI2C(byte address, int8_t num, byte _buff[]) {
   int8 buffer[7];
   buffer[0]=(int8)(ADXL345_COMMAND_BIT | address); 
   i2c_transfer(ADXL345_stream, ADXL345_DEVICE<<1, buffer, 1, buffer, num);    
   for (int8 i=0; i<=num; ++i)
   {
      _buff[i]=buffer[i];
   }
}



/*************************** RANGE SETTING **************************/
/*          ACCEPTABLE VALUES: 2g, 4g, 8g, 16g ~ GET & SET          */
void getRangeSetting(byte* rangeSetting) {
   byte _b;
   readFrom(ADXL345_DATA_FORMAT, 1, &_b);
   *rangeSetting = _b & 0B00000011;
}

void setRangeSetting(int16_t val) {
   byte _s;
   byte _b;

   switch (val) {
      case 2:
         _s = 0B00000000;
         break;
      case 4:
         _s = 0B00000001;
         break;
      case 8:
         _s = 0B00000010;
         break;
      case 16:
         _s = 0B00000011;
         break;
      default:
         _s = 0B00000000;
   }
   readFrom(ADXL345_DATA_FORMAT, 1, &_b);
   _s |= (_b & 0B11101100);
   writeTo(ADXL345_DATA_FORMAT, _s);
}

/*************************** SELF_TEST BIT **************************/
/*                            ~ GET & SET                           */
boolean getSelfTestBit() {
   return getRegisterBit(ADXL345_DATA_FORMAT, 7);
}

// If Set (1) Self-Test Applied. Electrostatic Force exerted on the sensor
//  causing a shift in the output data.
// If Set (0) Self-Test Disabled.
void setSelfTestBit(boolean selfTestBit) {
   setRegisterBit(ADXL345_DATA_FORMAT, 7, selfTestBit);
}

/*********************** INT_INVERT BIT STATE ***********************/
/*                           ~ GET & SET                            */
boolean getInterruptLevelBit() {
   return getRegisterBit(ADXL345_DATA_FORMAT, 5);
}

// If Set (0) Sets the Interrupts to Active HIGH
// If Set (1) Sets the Interrupts to Active LOW
void setInterruptLevelBit(boolean interruptLevelBit) {
   setRegisterBit(ADXL345_DATA_FORMAT, 5, interruptLevelBit);
}

/************************* FULL_RES BIT STATE ***********************/
/*                           ~ GET & SET                            */
boolean getFullResBit() {
   return getRegisterBit(ADXL345_DATA_FORMAT, 3);
}

// If Set (1) Device is in Full Resolution Mode: Output Resolution Increase with G Range
//  Set by the Range Bits to Maintain a 4mg/LSB Scale Factor
// If Set (0) Device is in 10-bit Mode: Range Bits Determine Maximum G Range
//  And Scale Factor
void setFullResBit(boolean fullResBit) {
   setRegisterBit(ADXL345_DATA_FORMAT, 3, fullResBit);
}

/*************************** JUSTIFY BIT STATE **************************/
/*                           ~ GET & SET                            */
boolean getJustifyBit() {
   return getRegisterBit(ADXL345_DATA_FORMAT, 2);
}

// If Set (1) Selects the Left Justified Mode
// If Set (0) Selects Right Justified Mode with Sign Extension
void setJustifyBit(boolean justifyBit) {
   setRegisterBit(ADXL345_DATA_FORMAT, 2, justifyBit);
}

/*********************** THRESH_TAP BYTE VALUE **********************/
/*                          ~ SET & GET                             */
// Should Set Between 0 and 255
// Scale Factor is 62.5 mg/LSB
// A Value of 0 May Result in Undesirable Behavior
void setTapThreshold(int16_t tapThreshold) {
   tapThreshold = constrain(tapThreshold,0,255);
   byte _b = (byte) (tapThreshold);
   writeTo(ADXL345_THRESH_TAP, _b);
}

// Return Value Between 0 and 255
// Scale Factor is 62.5 mg/LSB
int16_t getTapThreshold() {
   byte _b;
   readFrom(ADXL345_THRESH_TAP, 1, &_b);
   return (int16_t) (_b);
}

/****************** GAIN FOR EACH AXIS IN Gs / COUNT *****************/
/*                           ~ SET & GET                            */
void setAxisGains(double *_gains){
   int16_t i;
   for(i = 0; i < 3; i++){
      gains[i] = _gains[i];
   }
}
void getAxisGains(double *_gains){
   int16_t i;
   for(i = 0; i < 3; i++){
      _gains[i] = gains[i];
   }
}

/********************* OFSX, OFSY and OFSZ BYTES ********************/
/*                           ~ SET & GET                            */
// OFSX, OFSY and OFSZ: User Offset Adjustments in Twos Complement Format
// Scale Factor of 15.6mg/LSB
void setAxisOffset(int16_t x, int16_t y, int16_t z) {
   writeTo(ADXL345_OFSX, (byte) (x));
   writeTo(ADXL345_OFSY, (byte) (y));
   writeTo(ADXL345_OFSZ, (byte) (z));
}

void getAxisOffset(int16_t* x, int16_t* y, int16_t*z) {
   byte _b;
   readFrom(ADXL345_OFSX, 1, &_b);
   *x = (int16_t) (_b);
   readFrom(ADXL345_OFSY, 1, &_b);
   *y = (int16_t) (_b);
   readFrom(ADXL345_OFSZ, 1, &_b);
   *z = (int16_t) (_b);
}

/****************************** DUR BYTE ****************************/
/*                           ~ SET & GET                            */
// DUR Byte: Contains an Unsigned Time Value Representing the Max Time
//  that an Event must be Above the THRESH_TAP Threshold to qualify
//  as a Tap Event
// The scale factor is 625s/LSB
// Value of 0 Disables the Tap/Double Tap Funcitons. Max value is 255.
void setTapDuration(int16_t tapDuration) {
   tapDuration = constrain(tapDuration,0,255);
   byte _b = (byte) (tapDuration);
   writeTo(ADXL345_DUR, _b);
}

int16_t getTapDuration() {
   byte _b;
   readFrom(ADXL345_DUR, 1, &_b);
   return (int16_t) (_b);
}

/************************** LATENT REGISTER *************************/
/*                           ~ SET & GET                            */
// Contains Unsigned Time Value Representing the Wait Time from the Detection
//  of a Tap Event to the Start of the Time Window (defined by the Window
//  Register) during which a possible Second Tap Even can be Detected.
// Scale Factor is 1.25ms/LSB.
// A Value of 0 Disables the Double Tap Function.
// It Accepts a Maximum Value of 255.
void setDoubleTapLatency(int16_t doubleTapLatency) {
   byte _b = (byte) (doubleTapLatency);
   writeTo(ADXL345_LATENT, _b);
}

int16_t getDoubleTapLatency() {
   byte _b;
   readFrom(ADXL345_LATENT, 1, &_b);
   return (int16_t) (_b);
}

/************************** WINDOW REGISTER *************************/
/*                           ~ SET & GET                            */
// Contains an Unsigned Time Value Representing the Amount of Time
//  After the Expiration of the Latency Time (determined by Latent register)
//  During which a Second Valid Tape can Begin.
// Scale Factor is 1.25ms/LSB.
// Value of 0 Disables the Double Tap Function.
// It Accepts a Maximum Value of 255.
void setDoubleTapWindow(int16_t doubleTapWindow) {
   doubleTapWindow = constrain(doubleTapWindow,0,255);
   byte _b = (byte) (doubleTapWindow);
   writeTo(ADXL345_WINDOW, _b);
}

int16_t getDoubleTapWindow() {
   byte _b;
   readFrom(ADXL345_WINDOW, 1, &_b);
   return (int16_t) (_b);
}

/*********************** THRESH_ACT REGISTER ************************/
/*                          ~ SET & GET                             */
// Holds the Threshold Value for Detecting Activity.
// Data Format is Unsigned, so the Magnitude of the Activity Event is Compared
//  with the Value is Compared with the Value in the THRESH_ACT Register.
// The Scale Factor is 62.5mg/LSB.
// Value of 0 may Result in Undesirable Behavior if the Activity Interrupt Enabled.
// It Accepts a Maximum Value of 255.
void setActivityThreshold(int16_t activityThreshold) {
   activityThreshold = constrain(activityThreshold,0,255);
   byte _b = (byte) (activityThreshold);
   writeTo(ADXL345_THRESH_ACT, _b);
}

// Gets the THRESH_ACT byte
int16_t getActivityThreshold() {
   byte _b;
   readFrom(ADXL345_THRESH_ACT, 1, &_b);
   return (int16_t) (_b);
}

/********************** THRESH_INACT REGISTER ***********************/
/*                          ~ SET & GET                             */
// Holds the Threshold Value for Detecting Inactivity.
// The Data Format is Unsigned, so the Magnitude of the INactivity Event is
//  Compared with the value in the THRESH_INACT Register.
// Scale Factor is 62.5mg/LSB.
// Value of 0 May Result in Undesirable Behavior if the Inactivity Interrupt Enabled.
// It Accepts a Maximum Value of 255.
void setInactivityThreshold(int16_t inactivityThreshold) {
   inactivityThreshold = constrain(inactivityThreshold,0,255);
   byte _b = (byte) (inactivityThreshold);
   writeTo(ADXL345_THRESH_INACT, _b);
}

int16_t getInactivityThreshold() {
   byte _b;
   readFrom(ADXL345_THRESH_INACT, 1, &_b);
   return (int16_t) (_b);
}

/*********************** TIME_INACT RESIGER *************************/
/*                          ~ SET & GET                             */
// Contains an Unsigned Time Value Representing the Amount of Time that
//  Acceleration must be Less Than the Value in the THRESH_INACT Register
//  for Inactivity to be Declared.
// Uses Filtered Output Data* unlike other Interrupt Functions
// Scale Factor is 1sec/LSB.
// Value Must Be Between 0 and 255.
void setTimeInactivity(int16_t timeInactivity) {
   timeInactivity = constrain(timeInactivity,0,255);
   byte _b = (byte) (timeInactivity);
   writeTo(ADXL345_TIME_INACT, _b);
}

int16_t getTimeInactivity() {
   byte _b;
   readFrom(ADXL345_TIME_INACT, 1, &_b);
   return (int16_t) (_b);
}

/*********************** THRESH_FF Register *************************/
/*                          ~ SET & GET                             */
// Holds the Threshold Value, in Unsigned Format, for Free-Fall Detection
// The Acceleration on all Axes is Compared with the Value in THRES_FF to
//  Determine if a Free-Fall Event Occurred.
// Scale Factor is 62.5mg/LSB.
// Value of 0 May Result in Undesirable Behavior if the Free-Fall interrupt Enabled.
// Accepts a Maximum Value of 255.
void setFreeFallThreshold(int16_t freeFallThreshold) {
   freeFallThreshold = constrain(freeFallThreshold,0,255);
   byte _b = (byte) (freeFallThreshold);
   writeTo(ADXL345_THRESH_FF, _b);
}

int16_t getFreeFallThreshold() {
   byte _b;
   readFrom(ADXL345_THRESH_FF, 1, &_b);
   return (int16_t) (_b);
}

/************************ TIME_FF Register **************************/
/*                          ~ SET & GET                             */
// Stores an Unsigned Time Value Representing the Minimum Time that the Value
//  of all Axes must be Less Than THRES_FF to Generate a Free-Fall Interrupt.
// Scale Factor is 5ms/LSB.
// Value of 0 May Result in Undesirable Behavior if the Free-Fall Interrupt Enabled.
// Accepts a Maximum Value of 255.
void setFreeFallDuration(int16_t freeFallDuration) {
   freeFallDuration = constrain(freeFallDuration,0,255);
   byte _b = (byte) (freeFallDuration);
   writeTo(ADXL345_TIME_FF, _b);
}

int16_t getFreeFallDuration() {
   byte _b;
   readFrom(ADXL345_TIME_FF, 1, &_b);
   return (int16_t) (_b);
}

/************************** ACTIVITY BITS ***************************/
/*                                                                  */
boolean isActivityXEnabled() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 6);
}
boolean isActivityYEnabled() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 5);
}
boolean isActivityZEnabled() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 4);
}
boolean isInactivityXEnabled() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 2);
}
boolean isInactivityYEnabled() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 1);
}
boolean isInactivityZEnabled() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 0);
}

void setActivityX(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 6, state);
}
void setActivityY(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 5, state);
}
void setActivityZ(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 4, state);
}
void setActivityXYZ(boolean stateX, boolean stateY, boolean stateZ) {
   setActivityX(stateX);
   setActivityY(stateY);
   setActivityZ(stateZ);
}
void setInactivityX(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 2, state);
}
void setInactivityY(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 1, state);
}
void setInactivityZ(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 0, state);
}
void setInactivityXYZ(boolean stateX, boolean stateY, boolean stateZ) {
   setInactivityX(stateX);
   setInactivityY(stateY);
   setInactivityZ(stateZ);
}

boolean isActivityAc() {
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 7);
}
boolean isInactivityAc(){
   return getRegisterBit(ADXL345_ACT_INACT_CTL, 3);
}

void setActivityAc(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 7, state);
}
void setInactivityAc(boolean state) {
   setRegisterBit(ADXL345_ACT_INACT_CTL, 3, state);
}

/************************* SUPPRESS BITS ****************************/
/*                                                                  */
boolean getSuppressBit(){
   return getRegisterBit(ADXL345_TAP_AXES, 3);
}
void setSuppressBit(boolean state) {
   setRegisterBit(ADXL345_TAP_AXES, 3, state);
}

/**************************** TAP BITS ******************************/
/*                                                                  */
boolean isTapDetectionOnX(){
   return getRegisterBit(ADXL345_TAP_AXES, 2);
}
void setTapDetectionOnX(boolean state) {
   setRegisterBit(ADXL345_TAP_AXES, 2, state);
}
boolean isTapDetectionOnY(){
   return getRegisterBit(ADXL345_TAP_AXES, 1);
}
void setTapDetectionOnY(boolean state) {
   setRegisterBit(ADXL345_TAP_AXES, 1, state);
}
boolean isTapDetectionOnZ(){
   return getRegisterBit(ADXL345_TAP_AXES, 0);
}
void setTapDetectionOnZ(boolean state) {
   setRegisterBit(ADXL345_TAP_AXES, 0, state);
}

void setTapDetectionOnXYZ(boolean stateX, boolean stateY, boolean stateZ) {
   setTapDetectionOnX(stateX);
   setTapDetectionOnY(stateY);
   setTapDetectionOnZ(stateZ);
}

boolean isActivitySourceOnX(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 6);
}
boolean isActivitySourceOnY(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 5);
}
boolean isActivitySourceOnZ(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 4);
}

boolean isTapSourceOnX(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 2);
}
boolean isTapSourceOnY(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 1);
}
boolean isTapSourceOnZ(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 0);
}

/*************************** ASLEEP BIT *****************************/
/*                                                                  */
boolean isAsleep(){
   return getRegisterBit(ADXL345_ACT_TAP_STATUS, 3);
}

/************************** LOW POWER BIT ***************************/
/*                                                                  */
boolean isLowPower(){
   return getRegisterBit(ADXL345_BW_RATE, 4);
}
void setLowPower(boolean state) {
   setRegisterBit(ADXL345_BW_RATE, 4, state);
}

/*************************** RATE BITS ******************************/
/*                                                                  */
double getRate(){
   byte _b;
   readFrom(ADXL345_BW_RATE, 1, &_b);
   _b &= 0B00001111;
   return (pow(2,((int16_t) _b)-6)) * 6.25;
}

void setRate(double rate){
   byte _b,_s;
   int16_t v = (int16_t) (rate / 6.25);
   int16_t r = 0;
   while (v >= 1)
   {
      r++;
   }
   if (r <= 9) {
      readFrom(ADXL345_BW_RATE, 1, &_b);
      _s = (byte) (r + 6) | (_b & 0B11110000);
      writeTo(ADXL345_BW_RATE, _s);
   }
}

/*************************** BANDWIDTH ******************************/
/*                          ~ SET & GET                             */
void set_bw(int8_t bw_code){
   if((bw_code < ADXL345_BW_0_05) || (bw_code > ADXL345_BW_1600)){
      status = false;
      error_code = ADXL345_BAD_ARG;
   }
   else{
      writeTo(ADXL345_BW_RATE, bw_code);
   }
}

byte get_bw_code(){
   byte bw_code;
   readFrom(ADXL345_BW_RATE, 1, &bw_code);
   return bw_code;
}

/************************* TRIGGER CHECK  ***************************/
/*                                                                  */
// Check if Action was Triggered in Interrupts
// Example triggered(interrupts, ADXL345_SINGLE_TAP);
boolean triggered(byte interrupts, int16_t mask){
   return ((interrupts >> mask) & 1);
}

/*
 ADXL345_DATA_READY
 ADXL345_SINGLE_TAP
 ADXL345_DOUBLE_TAP
 ADXL345_ACTIVITY
 ADXL345_INACTIVITY
 ADXL345_FREE_FALL
 ADXL345_WATERMARK
 ADXL345_OVERRUNY
 */


byte getInterruptSource() {
   byte _b;
   readFrom(ADXL345_INT_SOURCE, 1, &_b);
   return _b;
}

boolean getInterruptSource(byte interruptBit) {
   return getRegisterBit(ADXL345_INT_SOURCE,interruptBit);
}

boolean getInterruptMapping(byte interruptBit) {
   return getRegisterBit(ADXL345_INT_MAP,interruptBit);
}

/*********************** INTERRUPT MAPPING **************************/
/*         Set the Mapping of an Interrupt to pin1 or pin2          */
// eg: setInterruptMapping(ADXL345_INT_DOUBLE_TAP_BIT,ADXL345_INT2_PIN);
void setInterruptMapping(byte interruptBit, boolean interruptPin) {
   setRegisterBit(ADXL345_INT_MAP, interruptBit, interruptPin);
}

void setImportantInterruptMapping(int16_t single_tap, int16_t double_tap, int16_t free_fall, int16_t activity, int16_t inactivity) {
   if(single_tap == 1) {
      setInterruptMapping( ADXL345_INT_SINGLE_TAP_BIT,   ADXL345_INT1_PIN );}
   else if(single_tap == 2) {
      setInterruptMapping( ADXL345_INT_SINGLE_TAP_BIT,   ADXL345_INT2_PIN );}

   if(double_tap == 1) {
      setInterruptMapping( ADXL345_INT_DOUBLE_TAP_BIT,   ADXL345_INT1_PIN );}
   else if(double_tap == 2) {
      setInterruptMapping( ADXL345_INT_DOUBLE_TAP_BIT,   ADXL345_INT2_PIN );}

   if(free_fall == 1) {
      setInterruptMapping( ADXL345_INT_FREE_FALL_BIT,   ADXL345_INT1_PIN );}
   else if(free_fall == 2) {
      setInterruptMapping( ADXL345_INT_FREE_FALL_BIT,   ADXL345_INT2_PIN );}

   if(activity == 1) {
      setInterruptMapping( ADXL345_INT_ACTIVITY_BIT,   ADXL345_INT1_PIN );}
   else if(activity == 2) {
      setInterruptMapping( ADXL345_INT_ACTIVITY_BIT,   ADXL345_INT2_PIN );}

   if(inactivity == 1) {
      setInterruptMapping( ADXL345_INT_INACTIVITY_BIT,   ADXL345_INT1_PIN );}
   else if(inactivity == 2) {
      setInterruptMapping( ADXL345_INT_INACTIVITY_BIT,   ADXL345_INT2_PIN );}
}

boolean isInterruptEnabled(byte interruptBit) {
   return getRegisterBit(ADXL345_INT_ENABLE,interruptBit);
}

void setInterrupt(byte interruptBit, boolean state) {
   setRegisterBit(ADXL345_INT_ENABLE, interruptBit, state);
}

void singleTapINT(boolean status) {
   if(status) {
      setInterrupt( ADXL345_INT_SINGLE_TAP_BIT, 1);
   }
   else {
      setInterrupt( ADXL345_INT_SINGLE_TAP_BIT, 0);
   }
}
void doubleTapINT(boolean status) {
   if(status) {
      setInterrupt( ADXL345_INT_DOUBLE_TAP_BIT, 1);
   }
   else {
      setInterrupt( ADXL345_INT_DOUBLE_TAP_BIT, 0);
   }
}
void FreeFallINT(boolean status) {
   if(status) {
      setInterrupt( ADXL345_INT_FREE_FALL_BIT,  1);
   }
   else {
      setInterrupt( ADXL345_INT_FREE_FALL_BIT,  0);
   }
}
void ActivityINT(boolean status) {
   if(status) {
      setInterrupt( ADXL345_INT_ACTIVITY_BIT,   1);
   }
   else {
      setInterrupt( ADXL345_INT_ACTIVITY_BIT,   0);
   }
}
void InactivityINT(boolean status) {
   if(status) {
      setInterrupt( ADXL345_INT_INACTIVITY_BIT, 1);
   }
   else {
      setInterrupt( ADXL345_INT_INACTIVITY_BIT, 0);
   }
}

void setRegisterBit(byte regAdress, int16_t bitPos, boolean state) {
   byte _b;
   readFrom(regAdress, 1, &_b);
   if (state) {
      _b |= (1 << bitPos);  // Forces nth Bit of _b to 1. Other Bits Unchanged.
   }
   else {
      _b &= ~(1 << bitPos); // Forces nth Bit of _b to 0. Other Bits Unchanged.
   }
   writeTo(regAdress, _b);
}

boolean getRegisterBit(byte regAdress, int16_t bitPos) {
   byte _b;
   readFrom(regAdress, 1, &_b);
   return ((_b >> bitPos) & 1);
}



